#include <stdlib.h>

#define order 220
#define numberoftasks 6
#define grainsize 2500      



float crandom (int *seed);
void SubA(float a[order][order], float b[order][order],
          float c[order][order], float x[order],
          float y[order], int *n, int *outstanding,
          int *lockid, int *eventid, int *subseta,
          int *subsetb);
      


main( )
{
/*
;
;  This is a main function for the example 4. The example computes
;  the following equations in parallel:
;
;     [C]=[A][B]
;     {Y}=[C]{X}
;
;  where matrices [A] and [B] are of order (NxN), and vector {X} is known.
;  The example uses a parallel lock to handle the critical sections
;  in both equations. Task synchronization is handled by
;  a WaitFor function and a parallel event.
;
*/
   static int n, outstanding, lockid, eventid, subseta, subsetb;
   static int taskid[numberoftasks];
   int j4temp, k4temp, seed = 111111;
   static float a[order][order], b[order][order], c[order][order];
   static float x[order], y[order];

/*  generate data in a sequential procedure  */

   printf(">>> Generate data ...\n");
   n = order;
   for (j4temp = 0; j4temp < order; j4temp ++ )
      for (k4temp = 0; k4temp < order ; k4temp ++ )
          a[k4temp][j4temp]= crandom(&seed);
   for (j4temp = 0; j4temp < order; j4temp ++ )
      for (k4temp = 0; k4temp < order ; k4temp ++ )
          b[k4temp][j4temp]= crandom(&seed);
   for (j4temp = 0; j4temp < order; j4temp ++ )
      x[j4temp]= crandom(&seed);

/*  create parallel lock  */

   CreateParallelLock(&lockid);

/*  create parallel event  */

   CreateParallelEvent(&eventid);
/*
;
;  determine length of subset
;  (where "N*(2..." please see function SubA)
;
*/
   subseta = __min( grainsize / ( n * ( 2 * n + 1)),
                    order / numberoftasks);
   if( subseta == 0) subseta = 1;
   subsetb = __min( grainsize / ( 2 * n + 1),
                    order / numberoftasks);
   if( subsetb == 0) subsetb = 1;

/*  initialize  */

   outstanding = 0;
   j4temp =  numberoftasks;
   InitializeParallelEvent(&eventid, &j4temp);

/*  compute these two equation in parallel  */

   printf(">>> Compute in parallal ...\n");
   for (j4temp = 0; j4temp < numberoftasks - 1; j4temp ++)
        Dispatch(&taskid[j4temp], SubA, a,
                                        b,
                                        c,
                                        x,
                                        y,
                                        &n,
                                        &outstanding,
                                        &lockid,
                                        &eventid,
                                        &subseta,
                                        &subsetb,
                                        0);
   SubA(a, b, c, x, y, &n, &outstanding, &lockid, &eventid,
        &subseta, &subsetb);

/*  wait for the completion of these two equations  */

   WaitForAndTerminateAllTasks( );

/*  delete parallel lock  */

   DeleteParallelLock(&lockid);

/*  delete parallel event  */

   DeleteParallelEvent(&eventid);

/*  output  */

   printf("    y[  0]= %6.6f\n", y[0]);
   printf("    y[ 99]= %6.6f\n", y[99]);
   printf("    y[199]= %6.6f\n", y[199]);
}
void SubA(float a[order][order],
          float b[order][order],
          float c[order][order],
          float x[order],
          float y[order],
          int *n,
          int *outstanding,
          int *lockid,
          int *eventid,
          int *subseta,
          int *subsetb)
{
/*
;
;  function to compute vector {y} in parallel
;     1.a,b: known matrices, dimension[n,n]
;     2.c: matrix to compute, dimension[n,n]
;     3.x: known vector, dimension[n]
;     4.y: vector to compute, dimension[n]
;     5.n: order
;     6.outstanding: outstanding element that must be initialized
;                    to zero
;     7.lockid: lock identifier
;     8.eventid: identifier of parallel event
;     9.subseta: length of subset of [c]
;    10.subsetb: length of subset of {y}
;
*/

/*  private variables  */

      int j4temp$, k4temp$, l4temp$, m4temp$, yes$;
/*
;
;  [critical section] -- the first equation
;  dynamic distribution
;
*/
label_10:    
   LockonParallelLock(lockid);
   if( *outstanding > *n )
   {
       UnlockParallelLock(lockid);
       goto label_20;
   }
   j4temp$ = *outstanding;
   *outstanding = *outstanding + *subseta;
   UnlockParallelLock(lockid);
/*
;
;  [parallelism] -- the first equation
;  compute a partial rows of [c]
;
*/
   for ( k4temp$ = j4temp$;
         k4temp$ < __min( j4temp$ + *subseta, *n );
         k4temp$ ++ )
         for ( l4temp$ = 0; l4temp$ < *n; l4temp$ ++)
         {
            c[l4temp$][k4temp$] = 0.0;
            for (m4temp$ = 0; m4temp$ < *n; m4temp$ ++)
                 c[l4temp$][k4temp$] = c[l4temp$][k4temp$] +
                                       a[l4temp$][m4temp$] *
                                       b[m4temp$][k4temp$];
         }
/*  where the grain size = subseta*n*(2n+1)  */
      goto label_10;
/*
;
;  [task synchronization] -- the first equation
;  complete a cycle of parallel event
;
*/
label_20:    
   PostParallelEvent(eventid);
   IfCompleteParallelEvent( eventid, &yes$);


/*  [subsequently sequential fragment] -- the first equation  */

   if( yes$ == 0)
   {
      *outstanding = 0;
      CompleteParallelEvent(eventid);
   }
/*
;
;  [critical section] -- the second equation
;  dynamic distribution
;
*/
label_30:    
   LockonParallelLock(lockid);
   if( *outstanding > *n)
   {
      UnlockParallelLock(lockid);
      return;
   }
   j4temp$ = *outstanding;
   *outstanding = *outstanding + *subsetb;
   UnlockParallelLock(lockid);
/*
;
;  [parallelism]  -- the second equation
;  compute a partial elements of vector {y}
;
*/
   for ( k4temp$ = j4temp$;
         k4temp$ < __min( j4temp$ + *subsetb, *n);
         k4temp$ ++)
   {
      y[k4temp$] = 0.0;
      for (l4temp$ = 0; l4temp$ < *n; l4temp$ ++)
           y[k4temp$] = y[k4temp$] + c[k4temp$][l4temp$] * x[l4temp$];
   }
/*  where the grain size = subsetb*(2n+1)  */
   goto label_30;
}
#define m 2147483647
#define a 16807
float crandom(int *seed)
{
   double rtempa, rtempb;
   float rtemp;
   int itemp;
   rtempa = (double) *seed * a;
   itemp = (int) (rtempa / (double) m);
   rtempb = (double) m * (double) itemp;
   *seed = (int) (rtempa - rtempb);
   rtemp = (double) *seed / (double) m;
   return rtemp;
}
